#!/usr/bin/perl
#Rene Warren
#2008

use strict;
use Data::Dumper;
use Date::Manip;

my $jregex_file = "Jregex.config";
my $config = "alignment.config";

if($#ARGV<4){die "Usage: $0 <*.out> <*.contigs> <First 2 bases of C segment> <sample id> <run id> <adjust V end? (optional)> \n";}

if(! -e $ARGV[0]){die "Error: can't find file $ARGV[0] -- fatal\n";}
if(! -e $ARGV[1]){die "Error: can't find file $ARGV[1] -- fatal\n";}
if(! -e $jregex_file){die "Error: can't find file Jregex.config -- fatal\n";}
if(! -e $config){die "Error: can't find $config, make sure it is in your working directory -- fatal\n";}
my $ADJUST = $ARGV[5];

my($fk_sample__id, $fk_run__id)=($ARGV[3],$ARGV[4]);

### Need first bases of Constant segment for frame/translation computations
my $FIRSTC = uc($ARGV[2]);

my ($regex,$ordreg) = &readRegex($jregex_file);
my $conf = &readAlignConfig($config);
my $trbv = &readSegmentSequences($conf->{'seqcluster'},"TRBV");
my $trbc = &readSegmentSequences($conf->{'seqcluster'},"TRBC");
my $vdb = &readSeeds($ARGV[0],$conf);
&assessCDR3($vdb,$regex,$ordreg,$ARGV[1],$trbv);

exit;



#------------------------------------------------
# Getting configuration information
sub readAlignConfig{

   my $file = shift;

   my $conf;
   open(CONFIG,$file) || die "Can't read $file -- fatal.\n";
   my $concat="";
   while(<CONFIG>){
      chomp;
      my @a=split(/\;/);
      $conf->{$a[1]}=$a[2];
   #   $concat .= "$a[1]:$a[2] ";
   }
   close CONFIG;

   my $unixdate=`date \"\+%Y-%m-%d %H:%M:%S\"`;
   chomp($unixdate);
   
   open (RUN, ">insert_run.sql");
   print RUN "INSERT INTO run VALUES(\"\",\"$fk_sample__id\",\"$concat\",\"\",\"$unixdate\");";
   close RUN;

   return $conf;
}

#------------------------------------------------
sub readRegex{
   
   my $file = shift;
   my $regex;
   my @ordreg;
   open(IN,$file) || die "can't open $regex for reading.\n";
   while(<IN>){
      chomp;
      my @a=split(/\,/);
      $regex->{$a[0]}{'rest'} = $a[1];
      $regex->{$a[0]}{'name'} = $a[2];
      $regex->{$a[0]}{'deleted'} = $a[3];
      push @ordreg,$a[0];
   }
   close IN;
   return $regex,\@ordreg;
}
#------------------------------------------------
sub assessCDR3{

   my ($vdb, $regex, $ordreg, $contigs, $trbv) = @_;
   my @ordreg = @$ordreg;
   my $potpro = "seeds_with_potential_problems.txt";

   open(IN,$contigs) || die "Problem reading contig file $contigs\n";

   my ($sd,$total_seed,$total_contig) = (0,0,0);
   my ($trackAll,$t,$o,$s,$ok,$wr,$av);
   my $seed = "";
   my $tig = "";
    
   open (TIGLIST, ">iSSAKEtigList.txt") || die "can't open file for writing -- fatal.\n";
   open (DB, ">insert_ssakeContig.sql") || die "can't open file for writing -- fatal.\n";

   while(<IN>){

      chomp;
      #>contig100_11|size95|read106329|cov40293.09|seed:19730274_VweakAlign(-1)|seeddepth1
      #TATTTCTGTGCCAGCAGCAAAGCTAGAGGGAGAGGCgctgctgggcgtctgggcggaggactcctggttctgg

      if(/seed\:(.*)\_Vweak/){
          $total_contig++;
          $tig = $_;
          $seed = $1; 
          $sd=0;
          if(/seeddepth(\d+)/){$sd=$1;$total_seed+=$sd;}

      }elsif(/^([ACGT]*)([acgt]*)?$/ && defined $vdb->{$seed}){  # 06/03/09 rlw
         my $seq = $1 . $2;                                      # 06/03/09 rlw  
         my $seedseq = $1;                                       # 06/03/09 rlw

         my ($left,$right,$jname,$jrest,$jdel)=("","","","","");
         my (@varr,@vend,@vdel,@vframe);

         my $v_list = $vdb->{$seed};
         ### capture end of D. and D name
         foreach my $v (keys %$v_list){
            push @varr, $v;

            ### Adjust end of V?
            my $vbdel = 0;

            if($ADJUST){  # 06/03/09 rlw

               open(PRO,">$potpro") || die "can't open file $potpro for writing -- fatal.\n";

               my $truncated_seed = substr($seedseq, $v_list->{$v}{'align'}-15,15);
               my $vn = $1 if ($v=~/\|(\w+V\d+[^\|]*)\|/);
               my $truncated_v = substr($trbv->{$vn},length($trbv->{$vn}) - 30, 30);
   
               #print "$v\nV=$truncated_v -3\' seed:$seedseq -3\' truncated=$truncated_seed -3\'\n";

               SCAN:
               until($truncated_v =~ /$truncated_seed/){
                 #print "TRY:$vbdel $truncated_seed\n";
                 $vbdel++;
                 $truncated_seed = substr($seedseq, $v_list->{$v}{'align'}-15,15 - $vbdel);                  
                 if($vbdel == 3 || length($truncated_seed)<7){
                    #print "TOO MANY ATTEMPTS or word too short.\n";
                    $vbdel=0;
                    print PRO "$seedseq\n";
                    last SCAN;
                 }
               }
               close PRO;
               #print "FINAL:$vbdel $truncated_seed  EXO:$v_list->{$v}{'deleted'} HERE:$v_list->{$v}{'deleted'} + $vbdel  \n\n" if ($vbdel < 3);
            }
           

            $v_list->{$v}{'deleted'} = $v_list->{$v}{'deleted'} + $vbdel;
            $v_list->{$v}{'align'} = $v_list->{$v}{'align'} - $vbdel;
            $v_list->{$v}{'frame'} = $v_list->{$v}{'frame'} - $vbdel;
            my $getframe = $v_list->{$v}{'frame'} % 3; 
 
            push @vend, $v_list->{$v}{'align'};
            push @vdel, $v_list->{$v}{'deleted'};
            push @vframe, $getframe;
         }

         JREGEX:
         for my $rx (@ordreg){
            if($seq =~ /^(\S+)($rx)/i){
               ($left,$right)=($1,$2);          ### same as V, account for ALL possibilities (push...)
               $jname = $regex->{$rx}{'name'};
               $jrest = $regex->{$rx}{'rest'};
               $jdel = $regex->{$rx}{'deleted'};
               last JREGEX;  
            }
         }

         ### AMBIGUOUS ::
         if($left eq ""){
            #print "Insufficient information to figure out J or CDR3\n";
            if(length($seq) >= $conf->{'TIG_LENGTH_THRESHOLD'}){
               my $ct=0;
               my $cdr = "";
               my $uni = "";
               my $left = $seq; 

               print TIGLIST "$tig,ambiguous\n";

               foreach my $vname (@varr){
                  my $len = length($left)-$vend[$ct];

                  if($len == 0){
                     my $lastv = substr($left,$vend[$ct]-15,15);
                     $uni = uc($lastv); 
                  }elsif($len > 0){
                     $cdr = substr($left,$vend[$ct],$len);
                     my $lastv = substr($left,$vend[$ct]-15,15);
                     $uni = uc($lastv) . lc($cdr) . uc($right);
                  }
                  ### QC
                  #my $full = "N" x $vframe[$ct] . $uni;                 
                  #print "LA*** $full ***\n";
                  ###
                  $ct++;  ### V-segment array counter
               }
               if($uni ne ""){ ### no weird detection
                  $o->{$uni}{'depth'} += $sd;
               }
            }else{
               $s->{$seq}{'depth'} += $sd; 
            }
         }

         if($jname ne "" && $left ne ""){
            my ($ct,$tempokt)=(0,0);
            my $cdr = "";
            my $uni = ""; 
            my $uniA = "";
            my $best = "";
            my $clonotype = "";

            print TIGLIST "$tig,unambiguous\n";
            my @tot;

            VSEG:
            foreach my $vname (@varr){            ### Go through all possible V segments
               my $vn = $1 if ($vname=~/\|(\w+V\d+[^\|]*)\|/);
               my ($jn,$cnum) = ($1,$2) if ($jname=~/\|(\w+J(\d+)[^\|]*)\|/);
               my $len = length($left)-$vend[$ct];
               if($len == 0){
                  my $lastv = substr($left,$vend[$ct]-15,15);
                  my $allv = substr($left,0,$vend[$ct]);
                  $uni = uc($lastv) . uc($right);
                  $uniA = uc($allv) . uc($right);
               }elsif($len > 0){
                  $cdr = substr($left,$vend[$ct],$len);
                  my $lastv = substr($left,$vend[$ct]-15,15);
                  my $allv = substr($left,0,$vend[$ct]);
                  $cdr = lc($cdr);
                  $uni = uc($lastv) . $cdr . uc($right);
                  $uniA = uc($allv) . $cdr . uc($right);
               }
               next VSEG if ($uni eq "");

               ### QC
               my $full = "n" x $vframe[$ct] . $uni . $jrest . $FIRSTC;  
               my $partial = "n" x $vframe[$ct] . $uni;              

               my $fullA = $uniA . $jrest . $FIRSTC;
               my $rcFullA = reverseComplement($fullA);
               my $rcCDR3 = $2 if ($rcFullA =~ /^(\w{3})*\w{1}CC\w{4}CC([AG]AA(\w{3})*?[AG]CA)/);
               my $ntcdr3 = reverseComplement($rcCDR3);
               #$ntcdr3 = $1 if($fullA =~ /(TG[TC](\w{3})*?TT[TC])GG\w{4}GG\w{1}(\w{3})*$/);
               my $aacdr3 = &dna2protein($ntcdr3);
               $ntcdr3 = "x" if ($ntcdr3 eq "");
               $aacdr3 = "x" if ($aacdr3 eq "");
               
               ### full AA
               my $vvv = substr($trbv->{$vn},0,length($trbv->{$vn})-$vdel[$ct]);
               my $cn = "TRBC" . $cnum;
               my $complete_trbv = uc($vvv) . $cdr . uc($right) . $jrest . $trbc->{$cn};

               my $aa = &dna2protein($full);
               my $aap = &dna2protein($partial);
               my $complete_trbv_aa = &dna2protein($complete_trbv);

               $clonotype = $vn . "_" . $vdel[$ct] . "_" . $cdr . "_" . $jdel . "_" . $jn;
               my $frame_check=0;

               if($best eq ""){
                  $best = $uni;        
                  my @goodies = ($ntcdr3,$aacdr3,$partial,$aap,$full,$aa,$vn,$vdel[$ct],$vend[$ct],$vframe[$ct],$jn,$jdel,$jrest,$frame_check,$complete_trbv,$complete_trbv_aa,$cdr);      
                  $trackAll->{$uni}=\@goodies;
               }

               if($aa =~ /\*/ || $aa =~ /x$/){### A stop codon or an unkown base at JC junction is not good
                  #Stop codon may be due to error, bad mapping at end of V (exonerate) or J segment slide alternate. 
               }else{
                  $tempokt++;
                  $frame_check=1;
                  $best = $uni;
                  my @goodies = ($ntcdr3,$aacdr3,$partial,$aap,$full,$aa,$vn,$vdel[$ct],$vend[$ct],$vframe[$ct],$jn,$jdel,$jrest,$frame_check,$complete_trbv,$complete_trbv_aa,$cdr);
                  $trackAll->{$uni}=\@goodies;
               }
               my @vs = ("$uni", "$tig","$sd", "$ct","$clonotype","$ntcdr3","$aacdr3","$seq","$partial","$aap","$full","$aa","$vn","$vdel[$ct]","$vend[$ct]","$vframe[$ct]","$jn","$jdel","$jrest","$frame_check", "$complete_trbv", "$complete_trbv_aa","$cdr");               
               push @tot, \@vs;

               #print "$frame_check : $fullA : $ntcdr3 : $aacdr3\n";
 
               $ct++;  ### V-segment array counter
            }### loops thru V_segments

            if($best ne ""){ ### no weird detection
               $t->{$best}{'depth'} += $sd;
               my $possible_v = "";
               my $tmp_clonotype = "";
               foreach my $y (@tot){
                  my @z = @$y;
                  $possible_v = $z[12] . "_" . $z[13] . "_" . $z[22];
                  $tmp_clonotype = "_" . $z[17] . "_" . $z[16];
                  print DB "INSERT INTO ssakeContig VALUES(\"\",\"$fk_run__id\",\"$z[0]\",\"$z[1]\",\"$z[2]\",\"$z[3]\",\"$z[4]\",\"$z[5]\",\"$z[6]\",\"$z[7]\",\"$z[8]\",\"$z[9]\",\"$z[10]\",\"$z[11]\",\"$z[12]\",\"$z[13]\",\"$z[14]\",\"$z[15]\",\"$z[16]\",\"$z[17]\",\"$z[18]\",\"$z[19]\",\"$z[20]\",\"$z[21]\")\;";
                  $av->{$best}{$z[12]}++;
               }
               my $best_clonotype = "";
               my $poss_v = $av->{$best};
               $t->{$best}{'v_possible'} = keys (%$poss_v);
               if($t->{$best}{'v_possible'} == 1){
                  $best_clonotype = $possible_v . $tmp_clonotype;
               }else{
                  $best_clonotype = "*" . $tmp_clonotype;
               }
               $t->{$best}{'clonotype'} = $best_clonotype;
               if($tempokt){$ok->{$best}{'depth'} += $sd;}else{$wr->{$best}{'depth'} += $sd;} ### count aa validity of any combo.
            }
         }

         @vend = ();
         @varr = ();

      }#end fasta block
   }### file handle
  
   close DB; 
   close IN;
   close TIGLIST;

   #--------------------------CONTIG OUTPUT--------------------------

   ### UNAMBIGUOUS
   my ($c,$unitig_depth,$unila_depth,$unisa_depth,$unito_depth,$unitw_depth)=(0,0,0,0,0,0);

   ($unitig_depth,$c) = &write2File($t,"unambiguousContig.fa",$c,"tig");
   ($unila_depth,$c) = &write2File($o, "longAmbiguousContig.fa",$c,"unknowntigLong");
   ($unisa_depth,$c) = &write2File($s, "shortAmbiguousContig.fa",$c,"unknowntigShort");
   ($unito_depth,$c) = &write2File($ok, "unambiguousContigTranslateOK.fa",$c,"tigTranslateOK");
   ($unitw_depth,$c) = &write2File($wr, "unambiguousContigTranslateWrong.fa",$c,"tigTranslateWrong");

   ##### SUMMARY #####
   my $unitig = keys (%$t);
   my $unila = keys (%$o);
   my $unisa = keys (%$s);
   my $uniok = keys (%$ok);
   my $uniwr = keys (%$wr);

   my $ambiguous_unique = $unila + $unisa;
   my $ambiguous_many = $unila_depth + $unisa_depth; 

   open (TIG,">insert_contig.sql");
   foreach my $ut (sort {$t->{$b}{'depth'}<=>$t->{$a}{'depth'}} keys %$t){
      my $l = $trackAll->{$ut};
      my @l = @$l;

      print TIG "INSERT INTO contig VALUES(\"\",\"$fk_run__id\",\"$ut\",\"$t->{$ut}{'depth'}\",\"$t->{$ut}{'clonotype'}\",\"$l[0]\",\"$l[1]\",\"$l[2]\",\"$l[3]\",\"$l[4]\",\"$l[5]\",\"$l[6]\",\"$l[7]\",\"$l[8]\",\"$l[9]\",\"$l[10]\",\"$l[11]\",\"$l[12]\",\"$l[13]\",\"$t->{$ut}{'v_possible'}\",\"$l[14]\",\"$l[15]\");\n";
   }
   close TIG;
   open (SUM,">insert_summary.sql");
   print SUM "INSERT INTO summary VALUES(\"\",\"$fk_run__id\",\"$total_contig\",\"$total_seed\",\"$unitig\",\"$unitig_depth\",\"$uniok\",\"$unito_depth\",\"$uniwr\",\"$unitw_depth\",\"$ambiguous_unique\",\"$ambiguous_many\",\"$unisa\",\"$unisa_depth\",\"$unila\",\"$unila_depth\")";
   close SUM;
}

#------------------------------------------------
sub write2File{

   my ($hash, $file, $counter, $head) = @_;

   my $unidepth = 0;

   open (FA, ">$file") || die "Can't open $file for writing.\n";
   foreach my $seq (sort {$hash->{$b}{'depth'}<=>$hash->{$a}{'depth'}} keys %$hash){
      $counter++;
      print FA ">$head-$counter-$hash->{$seq}{'depth'}\n$seq\n";
      $unidepth += $hash->{$seq}{'depth'};
   }
   close FA;

   return $unidepth,$counter;
}

#------------------------------------------------
sub readSegmentSequences{

   my ($file,$tag) = @_;

   my $hash;
   my $name = "na";

   open(IN,$file)  || die "Can't open $file for reading -- fatal.\n";
   while(<IN>){
      chomp;
      if(/\>(\S+)/){
         if($1=~/($tag\d+\-?\d?)/){
            $name=$1;
         }else{
            $name="na";
         }
      }elsif(/^([ACGT]*)/i){
         $hash->{$name} = $hash->{$name} . $1 if($name ne "na");
#         print "$name\t $hash->{$name}\n";
      }
   }
   close IN;
   return $hash;
}

#------------------------------------------------
sub readSeeds{

   my ($allout, $conf) = @_;

   open(IN,$allout);
   ### Warning, will put lots of IDs in memory
   my $track; 

   while(<IN>){
      chomp;

      my @a=split(/\^/);

      if($a[3] > 0){### captures exonerate one line result

         if($a[6]=~/V_segment/ && ($a[1] eq "+" && ($a[3]-$a[5]) >= $conf->{'V_BASE_MISSING'} && $a[9] >= ($a[7] - $conf->{'V_BASE_FROM_EDGE'}))){  ### Seeds edge of V
            $track->{$a[2]}{$a[6]}{'start'}= $a[4];
            $track->{$a[2]}{$a[6]}{'align'} = $a[5];           #tracks portion of seed aligned to V
            ###     read#  vtype#  = read align end
            $track->{$a[2]}{$a[6]}{'deleted'} = ($a[7]-$a[9]); #tracks deleted bases
            $track->{$a[2]}{$a[6]}{'frame'} = $a[9];           #use to track frame
         }

         if($a[6]=~/V_segment/ && ($a[1] eq "-" && $a[4] >= $conf->{'V_BASE_MISSING'} && $a[8] >= ($a[7] - $conf->{'V_BASE_FROM_EDGE'}) ) ) {  ### Seeds edge of V
            $track->{$a[2]}{$a[6]}{'start'}= ($a[3] - $a[5]);
            $track->{$a[2]}{$a[6]}{'align'} = ($a[3] - $a[4]);   #tracks portion of seed aligned to V
            $track->{$a[2]}{$a[6]}{'deleted'} = ($a[7] - $a[8]); #tracks deleted bases
            $track->{$a[2]}{$a[6]}{'frame'} = $a[8];       #use to track frame
         }

         #  0    1 2  3  4  5                          6                          7   8   9  10  11 12 13
         #100.00^+^43^36^0^32^ENSG00000211739|ENST00000390386|TRBV12-3|V_segment^347^315^347^160^32^32^0
         #94.44^-^39600127^36^0^36^ENSG00000211730|ENST00000390377|TRBV7-7|V_segment^347^280^244^162^36^34^2
         #96.88^-^39600115^36^0^32^ENSG00000211756|ENST00000390403|TRBJ1-1|J_segment^48^35^3^151^32^31^1

      }
   }
   close IN;

   return $track;
}

#---------------------------------------------------
sub dna2protein{
# converts the plain nucleotide sequence to amino acid sequence

    my $seq=shift;
    my @temp_array=split(//,$seq);

    my %codon2aa;

    $codon2aa{"TTT"}= "F"; $codon2aa{"TTC"}= "F"; $codon2aa{"TTA"}="L";
    $codon2aa{"TTG"}= "L"; $codon2aa{"TCT"}= "S"; $codon2aa{"TCC"}= "S";
    $codon2aa{"TCA"}= "S"; $codon2aa{"TCG"}= "S"; $codon2aa{"TAT"}= "Y";
    $codon2aa{"TAC"}= "Y"; $codon2aa{"TGT"}= "C"; $codon2aa{"TGC"}= "C";
    $codon2aa{"CTT"}= "L"; $codon2aa{"CTC"}= "L"; $codon2aa{"CTA"}= "L";
    $codon2aa{"CTG"}= "L"; $codon2aa{"CCT"}= "P"; $codon2aa{"CCC"}= "P";
    $codon2aa{"CCA"}= "P"; $codon2aa{"CCG"}= "P"; $codon2aa{"CAT"}= "H";
    $codon2aa{"CAC"}= "H"; $codon2aa{"CAA"}= "Q"; $codon2aa{"CAG"}= "Q";
    $codon2aa{"CGT"}= "R"; $codon2aa{"CGC"}= "R"; $codon2aa{"CGA"}= "R";
    $codon2aa{"CGG"}= "R"; $codon2aa{"ATT"}= "I"; $codon2aa{"ATC"}= "I";
    $codon2aa{"ATA"}= "I"; $codon2aa{"ATG"}= "M"; $codon2aa{"ACT"}= "T";
    $codon2aa{"ACC"}= "T"; $codon2aa{"ACA"}= "T"; $codon2aa{"ACG"}= "T";
    $codon2aa{"AAT"}= "N"; $codon2aa{"AAC"}= "N"; $codon2aa{"AAA"}= "K";
    $codon2aa{"AAG"}= "K"; $codon2aa{"AGT"}= "S"; $codon2aa{"AGC"}= "S";
    $codon2aa{"AGA"}= "R"; $codon2aa{"AGG"}= "R"; $codon2aa{"GTT"}= "V";
    $codon2aa{"GTC"}= "V"; $codon2aa{"GTA"}= "V"; $codon2aa{"GTG"}= "V";
    $codon2aa{"GCT"}= "A"; $codon2aa{"GCC"}= "A"; $codon2aa{"GCA"}= "A";
    $codon2aa{"GCG"}= "A"; $codon2aa{"GAT"}= "D"; $codon2aa{"GAC"}= "D";
    $codon2aa{"GAA"}= "E"; $codon2aa{"GAG"}= "E"; $codon2aa{"GGT"}= "G";
    $codon2aa{"GGC"}= "G"; $codon2aa{"GGA"}= "G"; $codon2aa{"GGG"}= "G";
    $codon2aa{"TGG"}= "W";
    $codon2aa{"TAG"}= "*";
    $codon2aa{"TGA"}= "*";
    $codon2aa{"TAA"}= "*";

    my $aa_seq="";

    for (my $pos=0; $pos<$#temp_array; $pos+=3){
        my $codon=$temp_array[$pos].$temp_array[$pos+1].$temp_array[$pos+2];
        if (defined $codon2aa{$codon}){
            $aa_seq .= $codon2aa{$codon};
        }elsif(! defined $codon2aa{$codon} && defined $codon2aa{uc($codon)}){
            $aa_seq .= lc($codon2aa{uc($codon)});
        }
        #undecidable nucleic acid
        else{
            $aa_seq.="x";
        }
    }
    return $aa_seq;
}

#-----------------------
sub reverseComplement{
   $_ = shift;
   tr/ATGCatgc/TACGtacg/;
   return (reverse());
}

