#!/usr/bin/perl
#Rene Warren, April 2008


use strict;
use Data::Dumper;

my $config = "alignment.config";
my $pipeline = "na";
my $mode = "serial";


if($#ARGV<2){die "Usage: $0 <query fasta> <base name for files> <serial/parallel> <pipeline name (optional) -clean,finish- >\n";}
if(! -e $config){die "Error: can't find $config, make sure it is in your working directory -- fatal\n";}
if(! -e $ARGV[0]){die "Error: can't find file $ARGV[0] -- fatal\n";}

### Getting configuration information
my $conf = &readAlignConfig($config);

$pipeline = $ARGV[3] if ($ARGV[3] ne "");
$mode = "parallel" if ($ARGV[2] ne "serial");
my $dir = `pwd`;
chomp($dir);
my $base_job_name = $ARGV[1];
my $TIME_TO_WAIT = $conf->{'wait'};
my $log = $dir . "/qstat.txt";

if($pipeline eq "clean"){
   removeDataFromNodes($dir, $conf->{'headnode'});
   exit;  
}


### Launching the first job
my $base_name = "SDB1";
my $screen = "seqcluster";
my $super_base_name = $base_job_name . $base_name;
my $jobdir = $dir . "/" . $screen;

my $sub1 = "POOL.fa";
my $sub2 = "SEEDS.fa";


eval{


if($pipeline ne "finish"){

   ###----------------------------------
   ### Split Original Input
   `rm -rf $jobdir`;
   `mkdir $jobdir`;
   my $first_file_input = &splitInput($ARGV[0],$jobdir,$conf->{'cluster_jobs1'});

   my $db = $conf->{$screen};
   if(! -e $db){die "*$db* doesn't exist..check your input.\n";}
   my $db_name = $1 if ($db=~/([^\/]*)$/);

   print "Generating Shell Scripts...\n";
   my $job_files=&makeSS($conf->{'aligner'}, $conf->{'aligner_options1'}, $first_file_input,$jobdir, $super_base_name,$base_name,$db_name,$db,$mode);

   ### Execute jobs
   if($mode eq "parallel"){
      &executeJobs($job_files, $jobdir, $super_base_name, $conf->{'headnode'}, $conf->{'qsub'}, $conf->{'resource'});
      ### Monitor/wait for jobs to finish
      &monitorJobs($super_base_name, $log, $conf->{'headnode'}, $conf->{'qstat'});
   }else{
      foreach my $runfile (@$job_files){
         `chmod 755 $runfile`;
         my $cmd = "bash $runfile";
         print "Will Execute: $cmd\n";
         system($cmd);
      } 
   }
}
#--------------------concat final hits into a single file

my $allout = $screen . ".out";
`cat $jobdir/$super_base_name*.out > $allout`;
die "File $allout is empty or missing; There might be some issues with $jobdir/$super_base_name* -- fatal\n" if (-z $allout);
&createFiles($ARGV[0],$allout,$sub2,$sub1,$conf);

};###end eval block


my $date = `date`;

if($@){
   my $message = $@;
   my $failure = "\nSomething went wrong running $0 $date\n$message\n";
   print $failure;
   print LOG $failure;
}else{
   my $success = "\nScript executed normally $date\n";
   print $success;
   print LOG $success;
}

exit;








#------------------------------------------------
sub removeDataFromNodes{

   my ($dir, $headnode, $qsub) = @_;
   
   my $jobdir = $dir . "/clean";
   `rm -rf $jobdir`;
   `mkdir $jobdir`;
   
   for(my $x=1;$x<=150;$x++){
      my $file = "cleanNodes" . $x . ".sh";
      open(CLEAN, ">$jobdir/$file");
      print CLEAN "#!/bin/sh\n";
      print CLEAN "#\$ -S /bin/sh\n";
      print CLEAN "#\$ -V\n";
      print CLEAN "rm -rf /tmp/*";
      close CLEAN; 
   }

   for(my $x=1;$x<=150;$x++){
      my $file = "cleanNodes" . $x . ".sh";
      my $qsub_cmd= "ssh $headnode 'export SGE_ROOT=/opt/sge;cd $jobdir;$qsub -pe ncpus 1 -N clean$x $file' ";
      system($qsub_cmd);
   }
}

#------------------------------------------------
sub createFiles{

   my ($original,$allout,$seeds,$nonseeds,$conf) = @_;

   open(IN,$allout);
   ### Warning, will put lots of IDs in memory
   my $seen;
   my $seed;
   my $wga;
   my $reject; 

   while(<IN>){
      chomp;

      my @a=split(/\^/);

      if($a[3] > 0){### captures exonerate one line result

         $seen->{$a[2]} = 1;

         #  0    1 2  3  4  5                          6                          7   8   9  10  11 12 13
         #100.00^+^43^36^0^32^ENSG00000211739|ENST00000390386|TRBV12-3|V_segment^347^315^347^160^32^32^0
         #94.44^-^39600127^36^0^36^ENSG00000211730|ENST00000390377|TRBV7-7|V_segment^347^280^244^162^36^34^2
         #96.88^-^39600115^36^0^32^ENSG00000211756|ENST00000390403|TRBJ1-1|J_segment^48^35^3^151^32^31^1

         my $id = ($a[12] / $a[3]) * 100;

         #### V

         if($a[6]=~/V_segment/ && ($a[1] eq "+" && ($a[3]-$a[5]) >= $conf->{'V_BASE_MISSING'} && $a[9] >= ($a[7] - $conf->{'V_BASE_FROM_EDGE'}))){  ### Seeds edge of V
            $seed->{$a[2]} = 1;
         }

         if($a[6]=~/V_segment/ && ($a[1] eq "-" && $a[4] >= $conf->{'V_BASE_MISSING'} && $a[8] >= ($a[7] - $conf->{'V_BASE_FROM_EDGE'}) ) ) {  ### Seeds edge of V
            $seed->{$a[2]} = -1;
         }

         #### SURE REJECT

         if($a[6]=~/V_segment/ && $id >= $conf->{'MINIMUM_IDENTITY'}  && (($a[1] eq "+" && $a[9] < ($a[7] - $conf->{'V_BASE_FROM_EDGE'} - $conf->{'V_BASE_BUFFER'})) || $a[1] eq "-" && $a[8] < ($a[7] - $conf->{'V_BASE_FROM_EDGE'} - $conf->{'V_BASE_BUFFER'}))) {
            $reject->{$a[2]} = 1;
         }

         if($a[6]=~/C_segment/ && $id >= $conf->{'MINIMUM_IDENTITY'}) {
            $reject->{$a[2]} = 1;
         }

         if($a[6]=~/JC[12]/ && $id >= $conf->{'MINIMUM_IDENTITY'}) {
            $reject->{$a[2]} = 1;
         }

         if($a[6]=~/J_segment/ && $id >= $conf->{'MINIMUM_IDENTITY'} && ($a[1] eq "+" && $a[8] > ($a[7] - $conf->{'J_BASE_FROM_EDGE'}) || $a[1] eq "-"  && $a[9] > ($a[7] - $conf->{'J_BASE_FROM_EDGE'}) ) ){   ### Seeds edge of J
            $reject->{$a[2]} = 1;
         }
      }
   }
   close IN;

   open(OUT1, ">$seeds");
   open(OUT2, ">$nonseeds");
   open(FA,$original);

   my $flag;
   while(<FA>){
      chomp;

      if(/\>(\S+)/){

         my $head = $1;

         if(! defined $seen->{$head}){  ## Either a really poor alignment, a contaminant, very bad reads or D junction
          
            $head .= "-Unaligned";
            print OUT2 ">$head\n";
            $flag=2;

         }else{

            if(! defined $reject->{$head}){
               my $tmp = $head . "-Other";
               print OUT2 ">$tmp\n";
               $flag=2;
            }

#            if(defined $wga->{$head}){
#               my $tmp = $head . "-Other";
#               print OUT2 ">$tmp\n";
#               $flag=2;
#            }

            if(defined $seed->{$head}){
               my $tmp = $head . "_VweakAlign(" . $seed->{$head} . ")" ;
               print OUT1 ">$tmp\n";
               $flag=$seed->{$head};
            }
            
         }

      }else{
         if($flag == 2){print OUT2 "$_\n";}elsif($flag == 1){print OUT1 "$_\n";print OUT2 "$_\n";}elsif($flag == -1){print OUT2 "$_\n";tr/ACTG/TGAC/;my $rc=reverse();print OUT1 "$rc\n";}
         $flag=0;
      }
   }
   close FA;
   close OUT1;
   close OUT2;

   return;
}


#------------------------------------------------
sub executeJobs{

   my ($job_file, $jobdir, $super_base_name, $headnode, $qsub, $resource) = @_;

   my $job_count=0;

   foreach my $job_file(@$job_file){
      $job_count++;
      my $subjobname=$super_base_name . "." . $job_count;
      my $qsub_cmd= "ssh $headnode 'export SGE_ROOT=/opt/sge;cd $jobdir;$qsub $resource -N $subjobname $job_file' "; 
      print "$qsub_cmd\n";
      system($qsub_cmd);
      sleep(1);
   }
}

#------------------------------------------------
#Check status, wait until all sub jobs are done before resuming script
#The system call returns 0 or 1 depending on success or failure, respectively (yes 0==success!)
#I am not aware of other ways to capture the output generated by programs called by system, but to redirect the output to a file.
sub monitorJobs{

   my ($super_base_name, $log, $headnode, $qstat) = @_;

   my $flag=1;
   my @status;

   while($flag){

      print "There are still some jobs queued/running...will wait $TIME_TO_WAIT sec. jobs are $super_base_name*\n";
      sleep($TIME_TO_WAIT);
      my $qstat_cmd="ssh $headnode 'export SGE_ROOT=/opt/sge;$qstat > $log'";
      system($qstat_cmd);

      #`cat $log1 $log2 > $log`;
      #sleep(2);
      open (LOG, $log) || die "Can't open $log.\n";
      @status=<LOG>;
      close LOG;

      $flag=0; #reset flag
      foreach my $status_line(@status){
          if($status_line=~/$super_base_name/i){  ###this will capture only jobs related to this project
            $flag=1;
            print "$status_line\n";
         }
      }
   }

   print "All sub jobs of run $super_base_name have finished.\n";
}

#------------------------------------------
sub makeSS{

   my ($exec, $options, $file_input,$dir,$super_base_name,$base_name,$db_name,$db,$mode)=@_;   

   my @job_files;
   
   my $job_id=1;
   foreach my $input (@$file_input){

      my $output;
      my $tmp_output;  
 
      my $title;
      chomp($input);      

         my $input_fn;

         if ($input=~/([^\/]*)$/){
            my $filename=$1;
            $output = $dir ."/". $super_base_name . "." . $job_id . "_" . $db_name .".out";
            $tmp_output =  $super_base_name . "." . $job_id . "_" . $db_name .".out";            

            $title = $filename . "_" . $db_name;
            $input_fn = $filename;
         }

         my $work_dir = "/tmp/" . $base_name . $job_id;
         my $work_all = $work_dir . "/*";
         my $new_in = $work_dir . "/" . $input_fn; 
         my $new_db = $work_dir . "/" . $db_name;
         
         my $job_file=$dir . "/" . $super_base_name . "." . $job_id . ".sh";

         push @job_files, $job_file;
         open (JOB, ">$job_file") or die "can't open $job_file for writing.\n";
         print JOB "#!/bin/sh\n";
         print JOB "#\$ -S /bin/sh\n";
         print JOB "#\$ -V\n";
         my $cmd1 = "rm -rf $work_all;rm -rf $work_dir;mkdir $work_dir;cp -rf $db $work_dir/.;cp -rf $input $new_in;cd $dir;$exec $options $new_in $new_db > $work_dir/$tmp_output;cp -rf $work_dir/$tmp_output $output;rm -rf $work_all;rm -rf $work_dir\n";
         my $cmd2 = "$exec $options $input $db > $output\n";
         print JOB $cmd1 if($mode eq 'parallel');
         print JOB $cmd2 if($mode eq 'serial');
         close JOB;
         $job_id++;
   }

   return \@job_files;
}

#------------------------------------------------
sub readAlignConfig{

   my $file = shift;
   
   my $conf;
   open(CONFIG,$file) || die "Can't read $file -- fatal.\n";
   
   while(<CONFIG>){
      chomp;
      my @a=split(/\;/);
      $conf->{$a[1]}=$a[2];
   }
   close CONFIG;

   return $conf;
}

#--------------------------------------------
sub splitInput{

   my ($fa,$dir,$numjobs) = @_;

   ### Defining the number of entries in each job file
   my $count = `grep -c ">" $fa`;
   chomp($count);
   my $entries = int($count / $numjobs);

   ### splitting input fasta file
   print "There are $count entries in your input fasta file $fa.  Will split into $numjobs files each having ~$entries fasta entries.\n";

   open (OUT,$fa);

   my ($ct,$mod)=(0,1);

   my $modfa = $fa . "." . $mod;

   open (IN, ">$dir/$modfa");

   my $head="";
   my $unique;
   my $lcr=0;
   my $dup = 0;
   my $tot = 0;

   while(<OUT>){
      chomp;
      if(/(\>\S+)/){
         $head=$1;
      }elsif(/([ACGTNX]*)/i){
         my $seq = $1;
         #if($seq=~/A{7}/i || $seq=~/T{7}/i || $seq=~/C{7}/i || $seq=~/G{7}/i){
         #   $lcr++;
         #}elsif(! defined $unique->{$seq}){

         #if(! defined $unique->{$seq}){### don't screen for LCR
         #   $unique->{$seq}++;
            $ct++;
            print IN "$head\n$seq\n" if ($ct <= $entries);

            if($ct > $entries){
               $tot+=$ct;
               $ct = 0;
               $mod++;
               $modfa = $fa . "." . $mod;
               close IN;
               open (IN, ">$dir/$modfa");
            }
         #}elsif(defined $unique->{$seq}){
         #   $dup++;
         #}
      }
   }
   $unique={};
   $tot+=$ct;

   close OUT;
   close IN;

   #print "Low complexity reads: $lcr\nExact duplicates: $dup\nTotal in split files: $tot\n";
   print "Total in split files: $tot\n";
   my @fof_ct=(1..$mod);
   my @file_input;
   foreach my $fc (@fof_ct){my $concat = $dir . "/" . $fa . "." . $fc;push @file_input, $concat;}

   return \@file_input;
}
